"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BigqueryQuery = void 0;
const shared_1 = require("@cubejs-backend/shared");
const BaseQuery_1 = require("./BaseQuery");
const BaseFilter_1 = require("./BaseFilter");
const GRANULARITY_TO_INTERVAL = {
    day: 'DAY',
    week: 'WEEK(MONDAY)',
    hour: 'HOUR',
    minute: 'MINUTE',
    second: 'SECOND',
    month: 'MONTH',
    quarter: 'QUARTER',
    year: 'YEAR'
};
class BigqueryFilter extends BaseFilter_1.BaseFilter {
    likeIgnoreCase(column, not, param, type) {
        const p = (!type || type === 'contains' || type === 'ends') ? '%' : '';
        const s = (!type || type === 'contains' || type === 'starts') ? '%' : '';
        return `LOWER(${column})${not ? ' NOT' : ''} LIKE CONCAT('${p}', LOWER(${this.allocateParam(param)}) , '${s}')`;
    }
    castParameter() {
        if (this.definition().type === 'boolean') {
            return 'CAST(? AS BOOL)';
        }
        else if (this.measure || this.definition().type === 'number') {
            // TODO here can be measure type of string actually
            return 'CAST(? AS FLOAT64)';
        }
        return '?';
    }
    castToString(sql) {
        return `CAST(${sql} as STRING)`;
    }
}
class BigqueryQuery extends BaseQuery_1.BaseQuery {
    castToString(sql) {
        return `CAST(${sql} as STRING)`;
    }
    convertTz(field) {
        return `DATETIME(${this.timeStampCast(field)}, '${this.timezone}')`;
    }
    timeStampCast(value) {
        return `TIMESTAMP(${value})`;
    }
    dateTimeCast(value) {
        return `DATETIME(TIMESTAMP(${value}))`;
    }
    escapeColumnName(name) {
        return `\`${name}\``;
    }
    timeGroupedColumn(granularity, dimension) {
        return `DATETIME_TRUNC(${dimension}, ${GRANULARITY_TO_INTERVAL[granularity]})`;
    }
    /**
     * Returns sql for source expression floored to timestamps aligned with
     * intervals relative to origin timestamp point.
     * BigQuery operates with whole intervals as is without measuring them in plain seconds.
     */
    dateBin(interval, source, origin) {
        const [intervalFormatted, timeUnit] = this.formatInterval(interval);
        const beginOfTime = this.dateTimeCast('\'1970-01-01T00:00:00\'');
        return `(${this.dateTimeCast(`'${origin}'`)} + INTERVAL ${intervalFormatted} *
      CAST(FLOOR(
        DATETIME_DIFF(${source}, ${this.dateTimeCast(`'${origin}'`)}, ${timeUnit}) /
        DATETIME_DIFF(${beginOfTime} + INTERVAL ${intervalFormatted}, ${beginOfTime}, ${timeUnit})
      ) AS INT64))`;
    }
    /**
     * The input interval with (possible) plural units, like "2 years", "3 months", "4 weeks", "5 days"...
     * will be converted to BigQuery dialect.
     * @see https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#interval_type
     * It returns a tuple of (formatted interval, timeUnit to use in datediff functions)
     */
    formatInterval(interval) {
        const intervalParsed = (0, shared_1.parseSqlInterval)(interval);
        const intKeys = Object.keys(intervalParsed).length;
        if (intervalParsed.year && intKeys === 1) {
            return [`${intervalParsed.year} YEAR`, 'YEAR'];
        }
        else if (intervalParsed.year && intervalParsed.month && intKeys === 2) {
            return [`'${intervalParsed.year}-${intervalParsed.month}' YEAR TO MONTH`, 'MONTH'];
        }
        else if (intervalParsed.year && intervalParsed.month && intervalParsed.day && intKeys === 3) {
            return [`'${intervalParsed.year}-${intervalParsed.month} ${intervalParsed.day}' YEAR TO DAY`, 'DAY'];
        }
        else if (intervalParsed.year && intervalParsed.month && intervalParsed.day && intervalParsed.hour && intKeys === 4) {
            return [`'${intervalParsed.year}-${intervalParsed.month} ${intervalParsed.day} ${intervalParsed.hour}' YEAR TO HOUR`, 'HOUR'];
        }
        else if (intervalParsed.year && intervalParsed.month && intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intKeys === 5) {
            return [`'${intervalParsed.year}-${intervalParsed.month} ${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}' YEAR TO MINUTE`, 'MINUTE'];
        }
        else if (intervalParsed.year && intervalParsed.month && intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intervalParsed.second && intKeys === 6) {
            return [`'${intervalParsed.year}-${intervalParsed.month} ${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}:${intervalParsed.second}' YEAR TO SECOND`, 'SECOND'];
        }
        else if (intervalParsed.quarter && intKeys === 1) {
            return [`${intervalParsed.quarter} QUARTER`, 'QUARTER'];
        }
        else if (intervalParsed.month && intKeys === 1) {
            return [`${intervalParsed.month} MONTH`, 'MONTH'];
        }
        else if (intervalParsed.month && intervalParsed.day && intKeys === 2) {
            return [`'${intervalParsed.month} ${intervalParsed.day}' MONTH TO DAY`, 'DAY'];
        }
        else if (intervalParsed.month && intervalParsed.day && intervalParsed.hour && intKeys === 3) {
            return [`'${intervalParsed.month} ${intervalParsed.day} ${intervalParsed.hour}' MONTH TO HOUR`, 'HOUR'];
        }
        else if (intervalParsed.month && intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intKeys === 4) {
            return [`'${intervalParsed.month} ${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}' MONTH TO MINUTE`, 'MINUTE'];
        }
        else if (intervalParsed.month && intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intervalParsed.second && intKeys === 5) {
            return [`'${intervalParsed.month} ${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}:${intervalParsed.second}' MONTH TO SECOND`, 'SECOND'];
        }
        else if (intervalParsed.week && intKeys === 1) {
            return [`${intervalParsed.week} WEEK`, 'DAY'];
        }
        else if (intervalParsed.day && intKeys === 1) {
            return [`${intervalParsed.day} DAY`, 'DAY'];
        }
        else if (intervalParsed.day && intervalParsed.hour && intKeys === 2) {
            return [`'${intervalParsed.day} ${intervalParsed.hour}' DAY TO HOUR`, 'HOUR'];
        }
        else if (intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intKeys === 3) {
            return [`'${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}' DAY TO MINUTE`, 'MINUTE'];
        }
        else if (intervalParsed.day && intervalParsed.hour && intervalParsed.minute && intervalParsed.second && intKeys === 4) {
            return [`'${intervalParsed.day} ${intervalParsed.hour}:${intervalParsed.minute}:${intervalParsed.second}' DAY TO SECOND`, 'SECOND'];
        }
        else if (intervalParsed.hour && intervalParsed.minute && intKeys === 2) {
            return [`'${intervalParsed.hour}:${intervalParsed.minute}' HOUR TO MINUTE`, 'MINUTE'];
        }
        else if (intervalParsed.hour && intervalParsed.minute && intervalParsed.second && intKeys === 3) {
            return [`'${intervalParsed.hour}:${intervalParsed.minute}:${intervalParsed.second}' HOUR TO SECOND`, 'SECOND'];
        }
        else if (intervalParsed.minute && intervalParsed.second && intKeys === 2) {
            return [`'${intervalParsed.minute}:${intervalParsed.second}' MINUTE TO SECOND`, 'SECOND'];
        }
        // No need to support microseconds.
        throw new Error(`Cannot transform interval expression "${interval}" to BigQuery dialect`);
    }
    intervalAndMinimalTimeUnit(interval) {
        return this.formatInterval(interval);
    }
    newFilter(filter) {
        return new BigqueryFilter(this, filter);
    }
    dateSeriesSql(timeDimension) {
        return `${timeDimension.dateSeriesAliasName()} AS (${this.seriesSql(timeDimension)})`;
    }
    seriesSql(timeDimension) {
        const values = timeDimension.timeSeries().map(([from, to]) => `select '${from}' f, '${to}' t`).join(' UNION ALL ');
        return `SELECT ${this.dateTimeCast('dates.f')} date_from, ${this.dateTimeCast('dates.t')} date_to FROM (${values}) AS dates`;
    }
    timestampFormat() {
        return 'YYYY-MM-DD[T]HH:mm:ss.SSSSSS[Z]';
    }
    timestampPrecision() {
        return 6;
    }
    overTimeSeriesSelect(cumulativeMeasures, dateSeriesSql, baseQuery, dateJoinConditionSql, baseQueryAlias) {
        const forSelect = this.overTimeSeriesForSelect(cumulativeMeasures);
        const outerSeriesAlias = this.cubeAlias('outer_series');
        const outerBase = this.cubeAlias('outer_base');
        const timeDimensionAlias = this.timeDimensions.map(d => d.aliasName()).filter(d => !!d)[0];
        const aliasesForSelect = this.timeDimensions.map(d => d.dateSeriesSelectColumn(outerSeriesAlias)).concat(this.dimensions.concat(cumulativeMeasures).map(s => s.aliasName())).filter(c => !!c).join(', ');
        const dateSeriesAlias = this.timeDimensions.map(d => `${d.dateSeriesAliasName()}`).filter(c => !!c)[0];
        return `
    WITH ${dateSeriesSql} SELECT ${aliasesForSelect} FROM
    ${dateSeriesAlias} ${outerSeriesAlias}
    LEFT JOIN (
      SELECT ${forSelect} FROM ${dateSeriesAlias}
      INNER JOIN (${baseQuery}) AS ${baseQueryAlias} ON ${dateJoinConditionSql}
      ${this.groupByClause()}
    ) AS ${outerBase} ON ${outerSeriesAlias}.${this.escapeColumnName('date_from')} = ${outerBase}.${timeDimensionAlias}
    `;
    }
    subtractInterval(date, interval) {
        return `DATETIME_SUB(${date}, INTERVAL ${this.formatInterval(interval)[0]})`;
    }
    addInterval(date, interval) {
        return `DATETIME_ADD(${date}, INTERVAL ${this.formatInterval(interval)[0]})`;
    }
    subtractTimestampInterval(date, interval) {
        const [intervalFormatted, timeUnit] = this.formatInterval(interval);
        if (['YEAR', 'MONTH', 'QUARTER'].includes(timeUnit)) {
            return this.timeStampCast(`DATETIME_SUB(DATETIME(${date}), INTERVAL ${intervalFormatted})`);
        }
        return `TIMESTAMP_SUB(${date}, INTERVAL ${intervalFormatted})`;
    }
    addTimestampInterval(date, interval) {
        const [intervalFormatted, timeUnit] = this.formatInterval(interval);
        if (['YEAR', 'MONTH', 'QUARTER'].includes(timeUnit)) {
            return this.timeStampCast(`DATETIME_ADD(DATETIME(${date}), INTERVAL ${intervalFormatted})`);
        }
        return `TIMESTAMP_ADD(${date}, INTERVAL ${intervalFormatted})`;
    }
    nowTimestampSql() {
        return 'CURRENT_TIMESTAMP()';
    }
    unixTimestampSql() {
        return `UNIX_SECONDS(${this.nowTimestampSql()})`;
    }
    // eslint-disable-next-line no-unused-vars
    preAggregationLoadSql(cube, preAggregation, tableName) {
        return this.preAggregationSql(cube, preAggregation);
    }
    hllInit(sql) {
        return `HLL_COUNT.INIT(${sql})`;
    }
    hllMerge(sql) {
        return `HLL_COUNT.MERGE(${sql})`;
    }
    countDistinctApprox(sql) {
        return `APPROX_COUNT_DISTINCT(${sql})`;
    }
    concatStringsSql(strings) {
        return `CONCAT(${strings.join(', ')})`;
    }
    defaultRefreshKeyRenewalThreshold() {
        return 120;
    }
    defaultEveryRefreshKey() {
        return {
            every: '2 minutes'
        };
    }
    sqlTemplates() {
        const templates = super.sqlTemplates();
        templates.quotes.identifiers = '`';
        templates.quotes.escape = '\\`';
        templates.functions.DATETRUNC = 'DATETIME_TRUNC(CAST({{ args[1] }} AS DATETIME), {% if date_part|upper == \'WEEK\' %}{{ \'WEEK(MONDAY)\' }}{% else %}{{ date_part }}{% endif %})';
        templates.functions.LOG = 'LOG({{ args_concat }}{% if args[1] is undefined %}, 10{% endif %})';
        templates.functions.BTRIM = 'TRIM({{ args_concat }})';
        templates.functions.STRPOS = 'STRPOS({{ args_concat }})';
        templates.functions.DATEDIFF = 'DATETIME_DIFF(CAST({{ args[2] }} AS DATETIME), CAST({{ args[1] }} AS DATETIME), {{ date_part }})';
        // DATEADD is being rewritten to DATE_ADD
        templates.functions.DATE_ADD = 'DATETIME_ADD(DATETIME({{ args[0] }}), INTERVAL {{ interval }} {{ date_part }})';
        templates.functions.CURRENTDATE = 'CURRENT_DATE';
        delete templates.functions.TO_CHAR;
        templates.expressions.binary = '{% if op == \'%\' %}MOD({{ left }}, {{ right }}){% else %}({{ left }} {{ op }} {{ right }}){% endif %}';
        templates.expressions.interval = 'INTERVAL {{ interval }}';
        templates.expressions.extract = 'EXTRACT({% if date_part == \'DOW\' %}DAYOFWEEK{% elif date_part == \'DOY\' %}DAYOFYEAR{% else %}{{ date_part }}{% endif %} FROM {{ expr }})';
        templates.expressions.timestamp_literal = 'DATETIME(TIMESTAMP(\'{{ value }}\'))';
        delete templates.expressions.ilike;
        delete templates.expressions.like_escape;
        templates.filters.like_pattern = 'CONCAT({% if start_wild %}\'%\'{% else %}\'\'{% endif %}, LOWER({{ value }}), {% if end_wild %}\'%\'{% else %}\'\'{% endif %})';
        templates.tesseract.ilike = 'LOWER({{ expr }}) {% if negated %}NOT {% endif %} LIKE {{ pattern }}';
        templates.types.boolean = 'BOOL';
        templates.types.float = 'FLOAT64';
        templates.types.double = 'FLOAT64';
        templates.types.decimal = 'BIGDECIMAL({{ precision }},{{ scale }})';
        templates.types.binary = 'BYTES';
        templates.expressions.cast_to_string = 'CAST({{ expr }} AS STRING)';
        templates.operators.is_not_distinct_from = 'IS NOT DISTINCT FROM';
        templates.join_types.full = 'FULL';
        templates.statements.time_series_select = 'SELECT DATETIME(TIMESTAMP(f)) date_from, DATETIME(TIMESTAMP(t)) date_to \n' +
            'FROM (\n' +
            '{% for time_item in seria  %}' +
            '    select \'{{ time_item[0] }}\' f, \'{{ time_item[1] }}\' t \n' +
            '{% if not loop.last %} UNION ALL\n{% endif %}' +
            '{% endfor %}' +
            ') AS dates';
        templates.statements.generated_time_series_select = 'SELECT DATETIME(d) AS date_from,\n' +
            'DATETIME_SUB(DATETIME_ADD(DATETIME(d),  INTERVAL {{ granularity }}), INTERVAL 1 MILLISECOND) AS date_to \n' +
            'FROM UNNEST(\n' +
            '{% if minimal_time_unit|upper in ["DAY", "WEEK", "MONTH", "QUARTER", "YEAR"] %}' +
            'GENERATE_DATE_ARRAY(DATE({{ start }}), DATE({{ end }}), INTERVAL {{ granularity }})\n' +
            '{% else %}' +
            'GENERATE_TIMESTAMP_ARRAY(TIMESTAMP({{ start }}), TIMESTAMP({{ end }}), INTERVAL {{ granularity }})\n' +
            '{% endif %}' +
            ') AS d';
        templates.statements.generated_time_series_with_cte_range_source = 'SELECT DATETIME(d) AS date_from,\n' +
            'DATETIME_SUB(DATETIME_ADD(DATETIME(d),  INTERVAL {{ granularity }}), INTERVAL 1 MILLISECOND) AS date_to \n' +
            'FROM {{ range_source }}, UNNEST(\n' +
            '{% if minimal_time_unit|upper in ["DAY", "WEEK", "MONTH", "QUARTER", "YEAR"] %}' +
            'GENERATE_DATE_ARRAY(DATE({{ range_source }}.{{ min_name }}), DATE({{ range_source }}.{{ max_name }}), INTERVAL {{ granularity }})\n' +
            '{% else %}' +
            'GENERATE_TIMESTAMP_ARRAY(TIMESTAMP({{ range_source }}.{{ min_name }}), TIMESTAMP({{ range_source }}.{{ max_name }}), INTERVAL {{ granularity }})\n' +
            '{% endif %}' +
            ') AS d';
        return templates;
    }
}
exports.BigqueryQuery = BigqueryQuery;
//# sourceMappingURL=BigqueryQuery.js.map